/*
"Mini Synth" Copyright (C) 2021  Kassen Oud

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA  02110-1301, USA.

This code owes a debt to the "Auduino" code by Peter Knight (
https://code.google.com/archive/p/tinkerit/wikis/Auduino.wiki ), for both the
overall structure and the "Grains" algorithm (licensed under the lgpl). Both
Jan Willem Hagenbeek and Rob Bothof influenced that code before I got it.

"Chord Organ" sound Copyright Eli A Zupke (Trainzack)
form; https://github.com/Trainzack/seeds-chord-organ
Licensed under the GPL-2

The rest is my own work.

To use this code you'll need the little synth that Jan Willem made for the
Lowlands and Dutch Modular Fest festivals. This should be available from
https://www.ginkosynthese.com/ by the time your read this. You may also want to
check https://www.mobilemodularsynth.com/ 

To upload this code, or your own versions of it, to the Arduino Nano on your
synth, you will need a mini USB cable and the Arduino IDE. This is available
from https://www.arduino.cc/ Instructions on how to do that are also found on
that site. Note that you should set the IDE to "Arduino Nano, old bootloader"
for uploads to work.

I hope this will be fun, useful, or educational to you; ideally all of the
above.
Yours, Kas.
*/

//include files
#include <avr/io.h>	//device specific in/out mapping
#include <avr/interrupt.h> //as we need an interrupt-based DSP loop
#include <avr/pgmspace.h> //to store our frequency table in progmem
#include "kas-button.h" //button de-bouncing


//definitions;

//sound constants, edit these to tweak the sound
const uint8_t grain_decay = 1;
const uint8_t grain2_decay = 2;
const uint8_t GRAIN_MOD_DECAY = 127;	//decay of mod envelope for the "grains" preset
const uint8_t WRAP_MOD_DECAY = 5;		//decay of mod envelope for the wave-wrapper preset
const uint8_t SAW_MOD_DECAY = 3;		//decay of mod envelope for the hoover preset
const uint8_t FM_MOD_DECAY = 4;			//decay of mod envelope for the fm preset
const uint8_t PD_MOD_DECAY = 5;			//decay of mod envelope for the phase distortion preset

// Map Analogue channels
#define FREQ_CONTROL			(3)			//Pitch input, marked 'CV1'. 
#define KEYBOARD				(0)			//Input from the keyboard (used for both CV and gate).
#define DECAY_CONTROL			(1)			//Decay knob, the middle one.
#define MOD_CONTROL				(2)			//Modulation knob, on the right.

//map Digital inputs;
#define KB_GATE					(10)		//Gate input, for external sequencing
#define OCT_UP_PIN				(8)			//"oct +" button. Sets the octave of the keyboard.
#define OCT_DOWN_PIN			(9)			//"oct -" button. Like above

// Changing these will also requires rewriting setup()
//    Output is on pin 11
#define PWM_PIN       11
#define PWM_VALUE     OCR2A
#define LED_PIN       13
#define LED_PORT      PORTB
#define LED_BIT       5
#define PWM_INTERRUPT TIMER2_OVF_vect

//constants

//table mapping keyboard notes to values in our tuning table
const uint16_t key_table[60] = {
  0,  17,  34,  51,  68,  85, 102, 119, 137, 154, 171, 188,
205, 222, 239, 256, 273, 290, 307, 324, 341, 358, 375, 393,
410, 427, 444, 461, 478, 495, 512, 529, 546, 563, 580, 597,
614, 631, 649, 666, 683, 700, 717, 734, 751, 768, 785, 802,
819, 836, 853, 870, 887, 905, 922, 939, 956, 973, 990, 1007};

//table mapping the release knob's value to an amount to be subtracted from the volume envelope
const uint16_t release_map[16] = {1024, 256, 128, 112,  96, 82, 64, 48,  36, 28, 24, 20,  18, 16, 12, 8};

const uint16_t max_int = 65535;
const uint16_t half_max_int = 32767;
const uint16_t quarter_max_int = 16383;

//Buttons. These are used to debounce on/off inputs
Button oct_up;
Button oct_down;
Button gate;

//variables relating to preset selection
uint8_t mode = 0;
const uint8_t grain_mode = 0;
const uint8_t tri_mode = 1;
const uint8_t hoover_mode = 2;
const uint8_t fm_mode = 3;
const uint8_t hs_mode = 4;
const uint8_t pd_mode = 5;
const uint8_t noise_mode = 6;
const uint8_t square_mode = 7;
const uint8_t ring_mode = 8;
const uint8_t wrap_mode = 9;
const uint8_t chord_mode = 10;
const uint8_t game_mode = 11;
const uint8_t num_modes = 12;

//Variables shared by all presets;
uint8_t mod_con_val = 0;
uint8_t dec_con_val = 0;
uint16_t master_phase = 0;
uint16_t m_phase_inc = 100;
uint16_t key_val = 0;
uint8_t octave = 2;
uint16_t found_key = 0;
uint16_t last_found_key = 0;
uint16_t pwmv = 0;
bool key_state;
bool keyb_was_last_trig = 0;
bool mode_switching = false;
bool gate_state = false;
bool sustain_state = false;
bool attacking = false;
uint16_t attack_delta = 255;
uint16_t decay = 0;
uint16_t env_value = 0;
uint16_t mod_env = 0;
uint16_t treated_mod_env = 0;
uint16_t mod_decay = GRAIN_MOD_DECAY;
uint8_t decay_counter = 0;
uint16_t filter_env = 0;
uint16_t filt_decay = 1;
uint16_t filt_sustain = 90 * 255;
uint8_t filt_val = 127;
uint32_t accumulator = 0;

//temporary values
uint8_t value = 0;
uint16_t output = 0;
uint16_t last_out = 0;
uint8_t mix_A = 0;
uint8_t mix_B = 0;
uint16_t out_A = 0;
uint16_t out_B = 0;
uint16_t target_pwmv = 0;
uint16_t pwmv_offset = 0;


//lfo variables and constants
uint8_t lfo_counter = 0;
uint8_t lfo_phase_inc = 7;				//LFO speed
uint16_t lfo_phase = 0;
uint16_t lfo_value = 0;
const uint8_t lfo_map[16] = {1, 2, 3, 4, 5, 6, 8, 12, 16, 24, 32, 48, 64, 82, 128, 255};
const uint8_t arp_map[16] = {16, 18, 22, 24, 28, 32, 40, 48, 56, 64, 73, 82, 105, 128, 190, 255};

//mode 0 (grains) phases
uint16_t grainPhaseInc = 0;
uint16_t grainPhaseAcc = 0;
uint16_t grainAmp = 0;
uint16_t grain2PhaseAcc = 0;
uint16_t grain2PhaseInc = 0;
uint16_t grain2Amp = 0;
uint16_t grain2val = 0;

//mode 1 (wave-wrapper) variables
bool flip = false;
uint16_t wrap = 1;				// amount of wave-folding
uint16_t drive = 32;			//amount of clipping for positive signal
uint16_t top_drive = 32;		//same for negative signal

//mode 2 (hoover) variables
bool cut = false;
bool last_cut = false;
uint16_t phase2 = 0;
uint16_t phase3 = 0;
uint16_t phase2_val = 0;
uint16_t phase3_val = 0;
uint32_t mod_value = 0;

//mode 3 (fm) variables
uint16_t carrier_phase = 0;
uint16_t carrier_value = 0;
uint16_t modulator_phase = 0;
uint16_t modulator_value = 0;
uint16_t fm_phase_inc = 1;
uint8_t fm_mod_depth = 100;

//mode 4 (hard sync) variables
uint16_t hs_mul = 1;

//mode 5 (phase distortion) variables
uint16_t treated_phase = 0;
uint8_t phase_dist_amount = 0;
uint16_t phase_mix_A = 0;
uint16_t phase_mix_B = 0;
uint16_t drive_factor = 0;
uint16_t o_drive = 255;
uint16_t d_mix_l = 0;
uint16_t d_mix_r = 0;

//mode 6 (noise) variables and function
uint8_t noise_count = 0;
uint8_t last_noise_count = 0;

//mode 7 (ring mod) variables
bool master_polarity = 0;
bool slave_polarity = 0;
uint16_t slave_offset = 1;
uint16_t slave_offset_measured = 1;

//mode 8 (wrap mode) variables
uint16_t wrap_factor = 64;
uint16_t neg_wrap_factor = 64;
uint16_t slave_phase = 0;
uint16_t master_value = 0;
uint16_t slave_value = 0;

//mode 9 (chord) variables
const uint8_t chord_ratio_exponent = 8;


// An array of chords, each number is a note with a ratio to the root of n/2^chord_ratio_exponent.
// In equal temperment (and with a ratio exponent of 8), the number of semitones above the root take on the following values:
//0:  256  1:   271  2:   287  3:   304  4:   323  5:   342  6:   362  7:   384  8:   406  9:   431  10:  456  11:  483  
//12: 512  13:  542  14:  575  15:  609  16:  645  17:  683  18:  724  19:  767  20:  813  21:  861  22:  912  23:  967
const uint8_t num_chords = 10;

const uint16_t chords[][4] = {
  { 256, 128, 256, 512},  // 8ves
  { 256, 256, 256, 256},  // unison
  { 256, 341, 455, 607},  // stacked perfect 4ths
  { 256, 384, 512, 683},  // sus4
  { 256, 384, 512, 640},  // major
  { 256, 384, 483, 640},  // major 7th
  { 256, 384, 455, 640},  // dominant 7th
  { 256, 384, 455, 614},  // minor 7th
  { 256, 384, 483, 614},  // minor major 7th
  { 256, 406, 512, 645},  // augmented
};

uint16_t chord_phase_1 = 0;
uint16_t chord_phase_2 = 0;
uint16_t chord_phase_3 = 0;
uint16_t chord_phase_4 = 0;

uint16_t chord_phase_inc_1 = 0;
uint16_t chord_phase_inc_2 = 0;
uint16_t chord_phase_inc_3 = 0;
uint16_t chord_phase_inc_4 = 0;

uint8_t current_chord = 0;

//mode 10 (game sound) variables
uint8_t chord_step = 0;
uint8_t chord_val = 0;

//"xorshift" noise generator because "rand()" costs quite a bit of CPU
//borrowed from http://www.arklyffe.com/main/2010/08/29/xorshift-pseudorandom-number-generator/
//As cryptography and related algorithms are generally released under an MIT licence I believe that
//borrowing this for a GPL-licensed program should be fine.
uint16_t rand_seed = 1;
uint16_t set_seed = 1;

uint16_t rng(void)
{
	rand_seed ^= (rand_seed << 13);
	rand_seed ^= (rand_seed >> 9);
	return rand_seed ^= (rand_seed << 7);
}

//table mapping CV values measured to phase increase per sample
const uint16_t freqTable[] PROGMEM = {
69,69,69,69,70,70,70,70,70,71,71,71,71,72,72,72,72,73,73,73,73,
74,74,74,74,75,75,75,75,76,76,76,76,77,77,77,77,78,78,78,78,79,
79,79,79,80,80,80,81,81,81,81,82,82,82,82,83,83,83,84,84,84,84,
85,85,85,86,86,86,86,87,87,87,88,88,88,88,89,89,89,90,90,90,91,
91,91,91,92,92,92,93,93,93,94,94,94,95,95,95,96,96,96,97,97,97,
97,98,98,98,99,99,99,100,100,100,101,101,101,102,102,102,103,103,104,104,104,
105,105,105,106,106,106,107,107,107,108,108,109,109,109,110,110,110,111,111,111,112,
112,113,113,113,114,114,114,115,115,116,116,116,117,117,118,118,118,119,119,120,120,
120,121,121,122,122,122,123,123,124,124,124,125,125,126,126,127,127,127,128,128,129,
129,130,130,130,131,131,132,132,133,133,134,134,134,135,135,136,136,137,137,138,138,
139,139,139,140,140,141,141,142,142,143,143,144,144,145,145,146,146,147,147,148,148,
149,149,150,150,151,151,152,152,153,153,154,154,155,155,156,156,157,157,158,158,159,
159,160,161,161,162,162,163,163,164,164,165,165,166,167,167,168,168,169,169,170,171,
171,172,172,173,173,174,175,175,176,176,177,178,178,179,179,180,181,181,182,182,183,
184,184,185,185,186,187,187,188,189,189,190,190,191,192,192,193,194,194,195,196,196,
197,198,198,199,200,200,201,202,202,203,204,204,205,206,206,207,208,208,209,210,211,
211,212,213,213,214,215,216,216,217,218,218,219,220,221,221,222,223,224,224,225,226,
227,227,228,229,230,231,231,232,233,234,234,235,236,237,238,238,239,240,241,242,242,
243,244,245,246,246,247,248,249,250,251,251,252,253,254,255,256,257,257,258,259,260,
261,262,263,264,264,265,266,267,268,269,270,271,272,273,273,274,275,276,277,278,279,
280,281,282,283,284,285,286,287,288,288,289,290,291,292,293,294,295,296,297,298,299,
300,301,302,303,304,305,306,307,308,310,311,312,313,314,315,316,317,318,319,320,321,
322,323,324,325,327,328,329,330,331,332,333,334,335,337,338,339,340,341,342,343,345,
346,347,348,349,350,352,353,354,355,356,357,359,360,361,362,363,365,366,367,368,370,
371,372,373,375,376,377,378,380,381,382,383,385,386,387,389,390,391,393,394,395,397,
398,399,401,402,403,405,406,407,409,410,411,413,414,416,417,418,420,421,423,424,425,
427,428,430,431,433,434,436,437,438,440,441,443,444,446,447,449,450,452,453,455,456,
458,460,461,463,464,466,467,469,470,472,474,475,477,478,480,482,483,485,486,488,490,
491,493,495,496,498,500,501,503,505,506,508,510,511,513,515,517,518,520,522,524,525,
527,529,531,532,534,536,538,540,541,543,545,547,549,551,552,554,556,558,560,562,564,
566,567,569,571,573,575,577,579,581,583,585,587,589,591,593,595,597,599,601,603,605,
607,609,611,613,615,617,619,621,623,625,627,630,632,634,636,638,640,642,644,647,649,
651,653,655,658,660,662,664,666,669,671,673,675,678,680,682,685,687,689,691,694,696,
698,701,703,705,708,710,713,715,717,720,722,725,727,729,732,734,737,739,742,744,747,
749,752,754,757,759,762,764,767,770,772,775,777,780,783,785,788,791,793,796,799,801,
804,807,809,812,815,817,820,823,826,828,831,834,837,840,842,845,848,851,854,857,860,
862,865,868,871,874,877,880,883,886,889,892,895,898,901,904,907,910,913,916,919,922,
925,928,932,935,938,941,944,947,950,954,957,960,963,966,970,973,976,979,983,986,989,
993,996,999,1003,1006,1009,1013,1016,1020,1023,1027,1030,1033,1037,1040,1044,1047,1051,1054,1058,1061,
1065,1069,1072,1076,1079,1083,1087,1090,1094,1098,1101,1105,1109,1112,1116,1120,1124,1127,1131,1135,1139,
1143,1146,1150,1154,1158,1162,1166,1170,1174,1178,1182,1186,1189,1193,1197,1201,1206,1210,1214,1218,1222,
1226,1230,1234,1238,1242,1247,1251,1255,1259,1263,1268,1272,1276,1280,1285,1289,1293,1298,1302,1306,1311,
1315,1320,1324,1328,1333,1337,1342,1346,1351,1355,1360,1365,1369,1374,1378,1383,1388,1392,1397,1402,1406,
1411,1416,1421,1425,1430,1435,1440,1444,1449,1454,1459,1464,1469,1474,1479,1484,1489,1494,1499,1504,1509,
1514,1519,1524,1529,1534,1539,1545,1550,1555,1560,1565,1571,1576,1581,1587,1592,1597,1603,1608,1613,1619,
1624,1630,1635,1641,1646,1652,1657,1663,1668,1674,1679,1685,1691,1696,1702,1708,1714,1719,1725,1731,1737,
1742,1748,1754,1760,1766,1772,1778,1784,1790,1796,1802,1808,1814,1820,1826,1832,1838,1845,1851,1857,1863,
1869,1876,1882,1888,1895,1901,1907,1914,1920,1927,1933,1940,1946,1953,1959,1966,1972,1979,1986,1992,1999,
2006,2012,2019,2026,2033,2040,2046,2053,2060,2067,2074,2081,2088,2095,2102,2109
};

//function to look values up, from the above table, as that table resides in program memory
uint16_t mapFreq(uint16_t input) {
  return pgm_read_word_near(freqTable + input);
}


//set up our Arduino at bootup
void setup() {
	// Set up PWM to 31.25kHz, phase accurate
	TCCR2A = _BV(COM2A1) | _BV(WGM20);
	TCCR2B = _BV(CS20);
	TIMSK2 = _BV(TOIE2);
	// configure the pins we use
	pinMode(PWM_PIN,OUTPUT);
	pinMode(LED_PIN,OUTPUT);
	pinMode(KB_GATE, INPUT_PULLUP);
	pinMode(OCT_UP_PIN, INPUT_PULLUP);
	pinMode(OCT_DOWN_PIN, INPUT_PULLUP);
	pinMode(KB_GATE, INPUT_PULLUP);
	//set up our button debouncing 
	gate.pin = KB_GATE;
	oct_up.pin = OCT_UP_PIN;
	oct_down.pin = OCT_DOWN_PIN;
}


//simple lowpass filter using two one-pole filters in series
uint16_t filter( uint16_t signal, uint16_t cutoff )
{
	cutoff *= cutoff;
	cutoff >>= 8;
	mix_B = cutoff;
	mix_A = 255 - mix_B;
	out_A = (out_A * mix_A) + (signal * mix_B);
	out_A = out_A >> 8;
	out_B = (out_B * mix_A) + (out_A * mix_B);
	out_B = out_B >> 8;
	return out_B;
}


//main loop. This deals with getting and processing our various inputs
void loop()
{
	//reading the internal keyboard.
	//This gets a bit involved as we use a single keyboard for both gate and CV
	//Read the keyboard voltage
	key_val = analogRead(KEYBOARD);
	//a value of over 1V means a key is held down
	if (key_val > (uint16_t)190  )
	{
		if (!key_state) //we detect a key has been pressed
		{
			key_state = true;
			//We need to keep track of where the trigger came from as the
			//external inputs behave differently from the internal keyboard
			keyb_was_last_trig = true;
			//then we need to figure out which key has been pressed
			//note that here we also deal with preset selection
			if (key_val > 987)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = PD_MOD_DECAY;
					mode = noise_mode;
				}
				else found_key = 11;						//B
			}
			else if (key_val > 918)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = PD_MOD_DECAY;
					mode = game_mode;
				}
				else found_key = 10;						//A'
			}
			else if (key_val > 848)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = PD_MOD_DECAY;
					mode = pd_mode;
				}
				else found_key = 9;							//A
			}
			else if (key_val > 778)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mode = chord_mode;
				}
				else found_key = 8;							//G'
			}
			else if (key_val > 708)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mode = hs_mode;
				}
				found_key = 7;								//G
			}
			else if (key_val > 638)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = FM_MOD_DECAY;
					mode = wrap_mode;
				}
				found_key = 6;								//F'
			}
			else if (key_val > 568)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = FM_MOD_DECAY;
					mode = fm_mode;
				}
				else found_key = 5;							//F
			}
			else if (key_val > 498)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = SAW_MOD_DECAY;
					lfo_phase_inc = 7;				//LFO speed
					mode = hoover_mode;
				}
				else found_key = 4;							//E
			}
			else if (key_val > 428)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = WRAP_MOD_DECAY;
					mode = ring_mode;
				}
				else found_key = 3;							//D'
			}
			else if (key_val > 358)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = WRAP_MOD_DECAY;
					mode = tri_mode;
				}
				else found_key = 2;							//D
			}
			else if (key_val > 289)
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = WRAP_MOD_DECAY;
					mode = square_mode;
				}
				else found_key = 1;							//C'
			}
			else	
			{
				if (oct_up.state && oct_down.state)
				{
					mode_switching = true;
					mod_decay = GRAIN_MOD_DECAY;
					mode = grain_mode;
				}
				else found_key = 0;							//C
			}
			//we only need to play a sound when we didn't use the key-press to select a preset
			if (!mode_switching)
			{
				//put our key in the correct octave
				found_key += (octave * 12);
				//look up the related frequency
				pwmv = key_table[ found_key ];
				last_found_key = found_key;
				//the filter is more open for higher notes
				filt_sustain = (90 + (pwmv >> 4)) * 255;
				//reset the phase in the case of the hoover sound
				if (mode == hoover_mode)
				{
					master_phase = 0;
					phase2 = 0;
					phase3 = 0;
				}
				//reset the arpegio for the game mode
				if (mode == game_mode)
				{
					lfo_phase = 0;
					lfo_counter = 0;
				}
				//turn on the Arduino's led
				LED_PORT |= 1 << LED_BIT;
				//reset the envelopes
				sustain_state = true;
				attacking = true;
				master_phase = max_int;
				mod_env = max_int;
				filter_env = max_int;
			}
		}
		//deal with glides at legato playing
		else
		{
			if (key_val > 987)			found_key = 11;		//B
			else if (key_val > 918)		found_key = 10;		//A'
			else if (key_val > 848)		found_key = 9;		//A
			else if (key_val > 778)		found_key = 8;		//G'
			else if (key_val > 708)		found_key = 7;		//G
			else if (key_val > 638)		found_key = 6;		//F' 
			else if (key_val > 568)		found_key = 5;		//F
			else if (key_val > 498)		found_key = 4;		//E
			else if (key_val > 428)		found_key = 3;		//D'
			else if (key_val > 358)		found_key = 2;		//D
			else if (key_val > 289)		found_key = 1;		//C'
			else						found_key = 0;		//C
			found_key += (octave * 12);
			//we only start gliding if we measure the same key twice in a row
			//this avoids a glitch where notes can sweep down once all keys are released
			//as in that case the voltage don't always return to 0 fast enough
			if (found_key == last_found_key) target_pwmv = key_table[ found_key ];
			last_found_key = found_key;
			//actual glide is as simple as it gets.
			if (target_pwmv > pwmv) pwmv++;
			else if (target_pwmv < pwmv) pwmv--;
		}
	}
	else
	{
		//note-off events
		if (key_state) 
		{
			if (!mode_switching) LED_PORT ^= 1 << LED_BIT;
			key_state = false;
			sustain_state = false;
			mode_switching = false;
		}
	}
	//we only read the external gate input when no keys are held on the keyboard
	if (!key_state) 
	{
		if( updateButton( &gate)) //input debouncing
		{
			//note-on detected
			if (gate.state) 
			{
				LED_PORT |= 1 << LED_BIT;
				gate_state = true;
				sustain_state = true;
				attacking = true;
				keyb_was_last_trig = false;
				master_phase = max_int;
				mod_env = max_int;
				filter_env = max_int;
				pwmv = analogRead(FREQ_CONTROL);
				filt_sustain = (90 + (pwmv >> 4)) * 255;
				if (mode == 3)
				{
					master_phase = 0;
					phase2 = 0;
					phase3 = 0;
				}
			}
			//note off
			else
			{
				LED_PORT ^= 1 << LED_BIT;
				gate_state = false;
				sustain_state = false;
			}
		}
		else
		{
			//no glide when using external control; use your modules if/when you want that.
			if (gate_state) pwmv = analogRead(FREQ_CONTROL);
		}
	}
	//when using the internal keyboard the leftmost knob has a range of an octave
	//as opposed to 5 octaves
	if (keyb_was_last_trig)
	{
		pwmv_offset =  analogRead(FREQ_CONTROL) >> 2;
		pwmv_offset *= 205;
		pwmv_offset = pwmv_offset >> 8;
		m_phase_inc = mapFreq( min( 1023, pwmv + pwmv_offset) );
	}
	else
	{
		m_phase_inc = mapFreq( pwmv );
	}

	if (updateButton ( &oct_up ))
	{
		if (oct_up.state && octave < 4) octave++;
	}
	if (updateButton ( &oct_down ))
	{
		if (oct_down.state && octave > 0) octave--;
	}

	switch (mode)
	{
		case grain_mode:
			grain2PhaseInc = mapFreq(analogRead(MOD_CONTROL)) * 4;
			//avoid fetching progmem in the DSP loop
			grainPhaseInc = max( 600, mapFreq(mod_env >> 6) * 2 );
			break;
		case tri_mode:
			wrap =  min( 64 + (analogRead(MOD_CONTROL) >> 1), 511);
			drive = min( 32 + (mod_env >> 7), 511);
			top_drive = max( 32, drive >>1);
			break;
		case hoover_mode:
			mod_value = analogRead(MOD_CONTROL) >> 2;	//8 bits
			treated_mod_env = (mod_env >> 8);
			treated_mod_env *= treated_mod_env;
			if (treated_mod_env & 0b1000000000000000) treated_mod_env = ~mod_env;
			mod_value *= (treated_mod_env >> 7);
			mod_value >>= 8;
			mod_value *= m_phase_inc;
			mod_value >>= 8;
			m_phase_inc += mod_value;
			break;
		case fm_mode:
			fm_phase_inc = m_phase_inc;
			accumulator = m_phase_inc * (analogRead(MOD_CONTROL) >> 2);
			accumulator >>= 8;
			fm_phase_inc += accumulator;
			break;
		case hs_mode:
			hs_mul = analogRead(MOD_CONTROL);
			break;
		case pd_mode:
			phase_dist_amount = analogRead(MOD_CONTROL) >> 2;
			break;
		case noise_mode:
			set_seed = analogRead(MOD_CONTROL) + 1;
			break;
		case square_mode:
			lfo_phase_inc = lfo_map[(uint16_t)analogRead(MOD_CONTROL) >> 6];
			break;
		case ring_mode:
			slave_offset_measured = (uint16_t)analogRead(MOD_CONTROL) << 2 ;
			break;
		case wrap_mode:
			wrap_factor = min( 64 + (uint16_t)analogRead(MOD_CONTROL) >>  1, 511);
			neg_wrap_factor = max(64, wrap_factor >> 1);
			break;
		case chord_mode:
		  // Here we figure out what the currently selected chord is based on the mod knob and mod CV value
		  // There might be a cleaner way to do this, but it works
		  current_chord = map((uint16_t)analogRead(MOD_CONTROL) << 1, 0, 2048, 0, num_chords);
		  current_chord = constrain(current_chord, 0, num_chords - 1);
		  
		  // Here we figure out how much to increment the phase of each note based on the currently selected chord
		  // The small constant offsets are to ensure each oscillator is slightly out of unison
		  chord_phase_inc_1 = (( (uint32_t)(m_phase_inc) * chords[current_chord][1]) >> chord_ratio_exponent);
		  chord_phase_inc_2 = (( (uint32_t)(m_phase_inc) * chords[current_chord][2]) >> chord_ratio_exponent) + 1;
		  chord_phase_inc_3 = (( (uint32_t)(m_phase_inc) * chords[current_chord][3]) >> chord_ratio_exponent) - 1;
		  chord_phase_inc_4 = (( (uint32_t)(m_phase_inc) * chords[current_chord][4]) >> chord_ratio_exponent) + 2;
		  
		  break;
		case game_mode:
			lfo_phase_inc = arp_map[(uint16_t)analogRead(MOD_CONTROL) >> 6];
			break;
	}
	decay = release_map[ (analogRead(DECAY_CONTROL) >> 6)];
}


void grainProcess()
{
	if (master_phase < m_phase_inc) {
		// Time to start the next grain
		grainPhaseAcc = 0;
		grain2PhaseAcc = 0;
		grainAmp = 0x7fff;
		grain2Amp = 0x7fff;
	}

	// Increment the phase of the grain oscillators
	grainPhaseAcc += grainPhaseInc;
	grain2PhaseAcc += grain2PhaseInc;

	// Convert phase into a triangle wave
	value = (grainPhaseAcc >> 7) & 0xff;
	if (grainPhaseAcc & 0x8000) value = ~value;
	// Multiply by current grain amplitude to get sample
	output = value * (grainAmp >> 8);

	// Repeat for second grain
	value = (grain2PhaseAcc >> 7) & 0xff;
	if (grain2PhaseAcc & 0x8000) value = ~value;
	grain2val = value * (grain2Amp >> 8);

	if ((max_int - output) < grain2val) output = max_int; //clipping before the int overflows
	else output += grain2val;

	// Make the grain amplitudes decay by a factor every sample (exponential decay)
	grainAmp -= (grainAmp >> 8) * grain_decay;
	grain2Amp -= (grain2Amp >> 8) * grain2_decay;
	output >>= 8;
}


void triProcess()
{
	flip = master_phase & 0b1000000000000000;
	//turn the phase acumulator into 4 up-ramps
	output = (master_phase &  0b0011111111111111);
	//get these into 7 bit range. "flip" will be the 8th bit of the output
	output = output >> 7;
	//invert the 2nd and 4th of our 4 up-ramps to create our triangle wave
	if (master_phase & 0b0100000000000000) output = (uint16_t)127 - output;

	//amplify the signal for the wave-wrapping
	output *= wrap;
	output = output >> 6;
	//detect whether folding will be more involved than inverting the range
	//from 128 to 254
	if (output & 0b1111111100000000) 
	{
		//values between 255 and 511 fold back past the "0" line
		if (output & 0b0000000100000000) flip = !flip;
		//mask out bits beyond which the process just repeats
		output &= 0b0000000011111111;
	}
	//actual folding
	if (output > (uint16_t)127)
	{
		output = 127 - (output & 0b0000000001111111);
	}

	//apply the signal gain for asymetrical clipping
	output *= flip?drive:top_drive;
	output = output >> 5;
	//clip
	if (output >  (uint16_t)127) output = (uint16_t)127;

	//turn our 7bit unipolar value into a 8bit bipolar one
	if ( flip ) output = (uint16_t)127 - output;
	else output += (uint16_t)127;
}

void hooverProcess()
{
	lfo_phase += lfo_phase_inc;
	if (lfo_phase & 0b1000000000000000) lfo_value = ~lfo_phase;
	else lfo_value = lfo_phase;

	if ((master_phase & 0b0111111111111111) < lfo_value) cut = true;
	else cut = false;
	if (master_phase < m_phase_inc)
	{
		output = 32768;
	}
	else if (cut != last_cut)
	{
		output = master_phase >> 1;
		last_cut = cut;
	}
	else if (cut)
	{
		output = (uint16_t)0;
	}
	else
	{
		output = master_phase;
	}
	output = output >> 7;
	//phase2
	phase2 += (m_phase_inc + 3);
	if ((phase2 & 0b0111111111111111) < lfo_value) cut = true;
	else cut = false;
	if (phase2 < (m_phase_inc + 3))
	{
		phase2_val = 32768;
	}
	else if (cut != last_cut)
	{
		phase2_val = phase2_val >> 1;
		last_cut = cut;
	}
	else if (cut)
	{
		phase2_val = (uint16_t)0;
	}
	else
	{
		phase2_val = phase2;
	}
	//phase3
	phase3 += (m_phase_inc - 2);
	if ((phase3 & 0b0111111111111111) < lfo_value) cut = true;
	else cut = false;
	if (phase3 < (m_phase_inc - 2))
	{
		phase3_val = 32768;
	}
	else if (cut != last_cut)
	{
		phase3_val = phase2_val >> 1;
		last_cut = cut;
	}
	else if (cut)
	{
		phase3_val = (uint16_t)0;
	}
	else
	{
		phase3_val = phase3;
	}
	output += (phase2_val >>8);
	output += (phase3_val >>8);
	output >>= 2;
}


//Simple FM synth based on approximating two sines, with one modulating the phase of the other.
//Nearly all fm implementations are actually phase modulation, with good reason.
void fmProcess()
{
	fm_mod_depth = max( 25, mod_env >> 8);
	carrier_phase = master_phase;
	modulator_phase += fm_phase_inc;
	//We turn an 16 bit up-ramp into 2 8 bit up-ramps
	modulator_value = modulator_phase;
	modulator_value <<= 1;
	modulator_value >>= 8;
	//multiply by the inverse to approximate 2 half sines
	modulator_value *= (~modulator_value) & 255;
	modulator_value >>= 6;
	modulator_value *= fm_mod_depth;
	//no need to reconstruct the half sines into a sine, we can just use the first half for
	//positive modulation, then subtract the 2nd half.
	if (modulator_phase & 0b1000000000000000) carrier_phase += modulator_value;
	else carrier_phase -= modulator_value;
	//repeat the same for the carrier
	carrier_value = carrier_phase;
	carrier_value <<= 1;
	carrier_value >>= 8;
	carrier_value *= (~carrier_value) & 255;
	//down to 7 bits; the last bit is for the sign
	carrier_value >>= 7;
	//this time we reconstruct the whole sine
	if (carrier_phase & 0b1000000000000000) output = 127 - carrier_value;
	else output = 127 + carrier_value;
}


//this strategy for implementing hard-sync is based on my conversations with Rob Hordijk
//it's fast, and when we overshoot the master phase's zero-crossing then that doesn't cause
//jitter in the carrier, which is a major advantage.
//Basically we multiply the master phase by some factor and let that overflow (here we fake the
//overflow as we need the resolution). This gives us a new saw wave with a frequency above the 
//master phase. Then we apply an inverse of the master phase ramp as a volume contour, to get
//rid is the discontinuity at the sync point
void hardSyncProsess()
{
	//fake 32 bit accumulator as we need the resolution
	accumulator = master_phase;
	//note that hs_mul is 10 bits, so this goes up by 2 bits, or a factor of 4, being 2 octaves.
	accumulator *= hs_mul;
	accumulator >>= 8;
	//going up an octave for the carrier is arbitrary, but if we don't then we lose a lot of
	//volume for lower settings of the modulation knob, so we just do it; as I see it this little
	//synth is about having fun.
	accumulator += master_phase;
	accumulator += master_phase;
	//there's the fake overflow I mentioned.
	accumulator &= max_int;
	//multiply by the inverse of the phase
	accumulator *= ~master_phase;
	accumulator >>= 16;
	//we're back into 16 bits, then down to the final 8 bit.
	output = accumulator;
	output >>= 8;
}

void phaseDistortionProcess()
{
	phase_mix_A = 255 * ~phase_dist_amount;
	phase_mix_A >>= 8;
	phase_mix_B = (master_phase >> 8) * phase_dist_amount;
	phase_mix_B >>= 8;
	phase_mix_B *= master_phase >> 8;
	phase_mix_B >>= 8;
	treated_phase = (phase_mix_A + phase_mix_B) * (master_phase >> 8);
	treated_phase += quarter_max_int;
	output = treated_phase;
	output <<= 1;
	output >>= 8;
	output *= (~output) & 255;
	output >>= 8;
	o_drive = max( 175, mod_env >> 8);
	drive_factor = output * o_drive;
	drive_factor >>= 8;
	d_mix_l = (output * drive_factor) >> 8;
	d_mix_r = (output * (~drive_factor & (uint16_t)255)) >> 4;
	output = (d_mix_l + d_mix_r) >> 3;
	if (treated_phase & 0b1000000000000000) output += 127;
	else output = 127 - output;
}


void noiseProcess()
{
	if (master_phase < m_phase_inc) rand_seed = set_seed;
	noise_count = master_phase >> 8;
	noise_count &= 7;
	if (noise_count != last_noise_count)
	{
		output = rng() >> 8;
		last_noise_count = noise_count;
	}
}


void squareProcess()
{
	lfo_counter++;
	if (! (lfo_counter & 0b00000111)) lfo_phase += lfo_phase_inc;
	if (lfo_phase & 0b1000000000000000) lfo_value = ~lfo_phase;
	else lfo_value = lfo_phase;
	lfo_value += quarter_max_int;

	if (master_phase  < lfo_value) cut = true;
	else cut = false;
	if (cut != last_cut)
	{
		output = 127;
		last_cut = cut;
	}
	else if (cut)
	{
		//we don't go from 0 to 255 as that'd make the square way louder than the other sounds
		output = (uint16_t)50;
	}
	else
	{
		output = (uint16_t)205;
	}
}

void ringProcess()
{
	if (slave_offset < slave_offset_measured) slave_offset++;
	else if (slave_offset > slave_offset_measured) slave_offset--;

	//turn master into a sine, or rather; a close approximation. The proper sine function
	//is slow and practical tables are lo-fi and/or slow as progmem is so slow
	master_value = master_phase;
	master_value <<= 1;
	master_value >>= 8;
	master_value *= (~master_value) & 255;
	master_value <<= 2;
	master_value >>= 8;
	master_polarity = master_phase & 0b1000000000000000;

	//calculate the slave phase in 32bit resolution
	//multiply the phase by a 10 bit number, then go down 8 bits again
	//the result will be a 2 octave increase max.
	//On top of that we say that we want the slave to at least run at
	//the master's phase's rate.
	//Finally we truncate back into 16 bit, as though the slave phase flowed over
	accumulator = master_phase;
	accumulator *= slave_offset;
	accumulator >>= 8;
	accumulator += master_phase;
	accumulator &= max_int;

	slave_phase = accumulator;

	//turn slave into a sine too, like we did with the master.
	slave_value = slave_phase;
	slave_value <<= 1;
	slave_value >>= 8;
	slave_value *= (~slave_value) & 255;
	slave_value <<= 2;
	slave_value >>= 8;
	slave_polarity = slave_phase & 0b1000000000000000;

	//multiply master and slave for the ring-mod,
	//taking care to invert the phase where needed
	master_value *= slave_value;
	master_value >>= 8;
	master_value >>= 1;
	master_polarity ^= slave_polarity;

	//only now do we make a bipolar signal
	if (master_polarity) output = 127 - master_value;
	else output = 128 + master_value;
}

void wrapProcess()
{
	//turn master into a sine, or rather; a close approximation. The proper sine function
	//is slow and practical tables are lo-fi and/or slow as progmem is so slow
	master_value = master_phase;
	master_value <<= 1;
	master_value >>= 8;
	master_value *= (~master_value) & 255;
	master_value <<= 2;
	master_value >>= 8;
	master_polarity = master_phase & 0b1000000000000000;

	master_value >>= 1;

	//wave-wrapping can result in frequency doubling, and we want to preserve bass
	//so we use asymmetrical wave-wrapping
	master_value *= (master_polarity?wrap_factor:neg_wrap_factor);

	//bit-shifting by 8 bits is way faster than by other amounts
	//so this is faster than >>= 6, as we know the topmost bits are 0
	master_value <<= 2;
	master_value >>= 8;

	//detect whether folding will be more involved than inverting the range
	//from 128 to 254
	if (master_value & 0b1111111100000000) 
	{
		//values between 255 and 511 fold back past the "0" line
		if (master_value & 0b0000000100000000) master_polarity = !master_polarity;
		//mask out bits beyond which the process just repeats
		master_value &= 0b0000000011111111;
	}
	//actual folding
	if (master_value > (uint16_t)127)
	{
		master_value = 127 - (master_value & 0b0000000001111111);
	}

	//only now do we make a bipolar signal
	if (master_polarity) output = 127 - master_value;
	else output = 128 + master_value;
}

void chordProcess()
{
  // This could be a loop over two arrays, but it's unrolled to save processing power, though I'm not sure it really needs it 
	chord_phase_1 += chord_phase_inc_1;
	chord_phase_2 += chord_phase_inc_2;
	chord_phase_3 += chord_phase_inc_3;
	chord_phase_4 += chord_phase_inc_4;

	// Add the phases to the output, creating four upramp sawtooth waves.
	// We bitshift each phase down to adjust the level.
	output = (chord_phase_1 >> 10) +
		(chord_phase_2 >> 10) +
		(chord_phase_3 >> 10) +
		(chord_phase_4 >> 10);
}

void gameProcess()
{
	lfo_counter++;
	if (! (lfo_counter & 0b00000111)) lfo_phase += lfo_phase_inc;
	chord_val = lfo_phase >> 8;

	if (chord_val > 223)		chord_step = 7;
	else if (chord_val > 191)	chord_step = 6;
	else if (chord_val > 159)	chord_step = 5;
	else if (chord_val > 127)	chord_step = 4;
	else if (chord_val > 95)	chord_step = 3;
	else if (chord_val > 63)	chord_step = 2;
	else if (chord_val > 31)	chord_step = 1;
	else						chord_step = 0;

	switch (chord_step)
	{
		case 0:		//tonic
			break;
		case 1:		//+3, major third, 5/4
			master_phase += m_phase_inc >> 2;
			break;
		case 2:		//+5, perfect fourth, 4/3
			accumulator = m_phase_inc * 85;
			accumulator >>= 8;
			master_phase += accumulator;
			break;
		case 3:		//+7 perfect fifth, 3:2
			master_phase += m_phase_inc >> 1;
			break;
		case 4:		//+12 octave
			master_phase += m_phase_inc;
			break;
		case 5:
			master_phase += m_phase_inc;
			master_phase += (m_phase_inc >> 1);
			break;
		case 6:
			master_phase += m_phase_inc;
			accumulator = m_phase_inc * 170;
			accumulator >>= 8;
			master_phase += accumulator;
			break;
		case 7:
			master_phase += m_phase_inc;
			master_phase += m_phase_inc;
			break;
		default:
			master_phase = 0;
	}
	cut = (master_phase & 0b1000000000000000)? true: false;
	if (cut != last_cut) output = 127;
	else if (cut) output = 255;
	else output = 0;
	last_cut = cut;
}

SIGNAL(PWM_INTERRUPT)
{
	output = 0;
	master_phase += m_phase_inc;
	decay_counter++;
	decay_counter &= 0b00000111;

	switch (mode)
	{
		case grain_mode:
			grainProcess();
			break;
		case tri_mode:
			 triProcess();
			 break;
		case hoover_mode:
			hooverProcess();
			break;
		case fm_mode:
			fmProcess();
			break;
		case hs_mode:
			hardSyncProsess();
			break;
		case pd_mode:
			phaseDistortionProcess();
			break;
		case noise_mode:
			noiseProcess();
			break;
		case square_mode:
			squareProcess();
			break;
		case ring_mode:
			ringProcess();
			break;
		case wrap_mode:
			wrapProcess();
			break;
		case chord_mode:
			chordProcess();
			break;
		case game_mode:
			gameProcess();
			break;
	}

	if (mod_env > mod_decay) mod_env -= mod_decay;
	else mod_env = 0;

	//enveloppe stuff
	if (attacking)
	{
		if (env_value <= (max_int - attack_delta))
		{
			env_value += attack_delta;
		}
		else
		{
			env_value = max_int;
			attacking = false;
		}
	}
	else if (sustain_state)
	{
		if (filter_env  > filt_sustain + filt_decay) filter_env -= filt_decay;
		else filter_env = filt_sustain;
		filt_val = filter_env >> 8;
	}
	else
	{
		if (!decay_counter)
		{
			if (env_value > decay) env_value -= decay;
			else env_value = 0;
		}
		filt_val = min(env_value, filter_env) >> 8;
	}

	//apply filter
	output = filter (output, filt_val );


	//apply envelope
	output *= (env_value >> 8);
	output >>= 8;

	// Output to PWM (this is faster than using analogWrite)  
	PWM_VALUE = output;
}
